# coding=utf-8
# Copyright 2021-present the HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Torch utilities for the Trainer class.
"""

import torch
from packaging import version


def _get_learning_rate(self):
    last_lr = (
        # backward compatibility for pytorch schedulers
        self.lr_scheduler.get_last_lr()[0]
        if version.parse(torch.__version__) >= version.parse("1.4")
        else self.lr_scheduler.get_lr()[0]
    )
    return last_lr
